"-- Чтобы можно было использовать их в типах объявим что эти классы будут объявлены
CLASS lcl_super_select DEFINITION DEFERRED.
"-- Класс шага ------------------------------------------------------------------------------------
CLASS lcl_super_select_step DEFINITION FINAL FRIENDS lcl_super_select.
  PUBLIC SECTION.
    "-- Установить имя подпрограмм шага
    METHODS set_formname
      IMPORTING
        iv_formname   TYPE any
      RETURNING
        VALUE(rr_ret) TYPE REF TO lcl_super_select_step.
    "-- Установить тип подпрограммы подпрограммы (ТАБЛИЧНЫЙ ТИП!)
    METHODS set_type
      IMPORTING
        iv_type       TYPE any
        iv_alias      TYPE any OPTIONAL " Псевдонимы полей
      RETURNING
        VALUE(rr_ret) TYPE REF TO lcl_super_select_step.
    "-- Добавить зависимость
    METHODS add_relation
      IMPORTING
        iv_relation   TYPE any
      RETURNING
        VALUE(rr_ret) TYPE REF TO lcl_super_select_step.
    "-- Изменить флаг фильтра
    METHODS set_filter
      IMPORTING
        iv_filter     TYPE wdy_boolean
      RETURNING
        VALUE(rr_ret) TYPE REF TO lcl_super_select_step.
    "-- Название шага
    METHODS get_title
      RETURNING
        VALUE(rv_ret) TYPE string.
    "-- Шаг активен?
    METHODS is_disable
      RETURNING
        VALUE(rv_ret) TYPE wdy_boolean.
    "-- Скрывать данный шаг при визуализации
    METHODS set_hide
      IMPORTING
        iv_hide       TYPE wdy_boolean
      RETURNING
        VALUE(rr_ret) TYPE REF TO lcl_super_select_step.
    "-- Запустить таймер с именем iv_name
    METHODS timer_start
      IMPORTING
        iv_name TYPE any
      .
    "-- Останавлитваться перед выполнением шага
    METHODS break_point_set
      IMPORTING
        iv_value      TYPE any
      RETURNING
        VALUE(rr_ret) TYPE REF TO lcl_super_select_step.
    "-- Останавлитваться перед выполнением шага?
    METHODS break_point_get
      RETURNING
        VALUE(rv_ret) TYPE wdy_boolean.
    "-- Остановить таймер с именем iv_name
    METHODS timer_end
      IMPORTING
        iv_name TYPE any
      .
    "-- Добавить поле
    METHODS add_field
      IMPORTING
        iv_name      TYPE any
        iv_type      TYPE any
        iv_aliasform TYPE any OPTIONAL " Имя поля в подпрограмме расчета
      .
    "-- Добавить поле для коррекции
    METHODS add_field_correction
      IMPORTING
        iv_name      TYPE any
        iv_aliasform TYPE any OPTIONAL " Имя поля в подпрограмме расчета
      .

  PRIVATE SECTION.
    "--
    TYPES tt_super_select2_step TYPE STANDARD TABLE OF REF TO lcl_super_select_step.
    "-- Искать в родительских шагах шаг ir_step
    METHODS _find_parent
      IMPORTING
        ir_step       TYPE REF TO lcl_super_select_step
        it_history    TYPE tt_super_select2_step OPTIONAL
      RETURNING
        VALUE(rr_ret) TYPE wdy_boolean.
    "-- Уникальный идентифкатор (для генерации подпрограммы подготовки выполенния продпрограммы шага)
    DATA mv_uid TYPE string.
    "-- Указатель на родительский запрос
    DATA mr_select TYPE REF TO lcl_super_select.
    "-- Название шага
    DATA mv_title TYPE string.
    "-- Имя подрограммы шага
    DATA mv_formname TYPE string.
    "-- Тип данных подпрограммы
    DATA mv_type TYPE string.
    "-- Применяет фильтр
    DATA mv_filter TYPE wdy_boolean VALUE abap_undefined.
    "-- Зависимости, от которых зависит расчет шага
    TYPES:
      BEGIN OF ts_relation,
        name      TYPE char40,
        aliasform TYPE string,
      END OF ts_relation.
    TYPES tt_relation TYPE STANDARD TABLE OF ts_relation.
    DATA mt_relations TYPE tt_relation.
    "-- Список полей
    TYPES:
      BEGIN OF ts_field,
        name      TYPE string,
        aliasform TYPE string,
        type      TYPE string,
      END OF ts_field.
    TYPES tt_field TYPE STANDARD TABLE OF ts_field.
    DATA mt_fields TYPE tt_field.
    "-- Имя поля для корректировки
    CONSTANTS mc_correction_fieldname TYPE fieldname VALUE '_UID_ROW'.
    "-- Список полей для коррекции
    TYPES:
      BEGIN OF ts_field_correction,
        name      TYPE string,
        aliasform TYPE string,
      END OF ts_field_correction.
    TYPES tt_field_correction TYPE STANDARD TABLE OF ts_field_correction.
    DATA mt_fields_correction TYPE tt_field_correction.
    "-- Родительские шаги
    DATA mt_parents TYPE tt_super_select2_step.
    "-- Флаг отключения шага
    DATA mv_disable TYPE wdy_boolean VALUE abap_undefined.
    "-- Таймеры
    TYPES:
      BEGIN OF ts_timer,
        name  TYPE string,
        start TYPE timestamp,
        end   TYPE timestamp,
        time  TYPE i,
      END OF ts_timer.
    TYPES tt_timer TYPE HASHED TABLE OF ts_timer WITH UNIQUE KEY name.
    DATA mt_timers TYPE tt_timer.
    "-- Останавливать программу перед выполнение программ
    DATA mv_break_point TYPE wdy_boolean VALUE abap_false.
    "-- Скрывать шаг при визуализации
    DATA mv_hide TYPE wdy_boolean VALUE abap_false.
ENDCLASS.
"-- Класс для формирования запросов ------------------------------------------------------------------------------------
CLASS lcl_super_select DEFINITION FINAL FRIENDS lcl_super_select_step.
  PUBLIC SECTION.
    "-- Строка кода
    TYPES tv_code TYPE c LENGTH 255.
    "-- Строки кода
    TYPES tt_code TYPE STANDARD TABLE OF tv_code.
    "-- Псевдонимы
    TYPES:
      BEGIN OF ts_item,
        left  TYPE string,
        right TYPE string,
      END OF ts_item.
    TYPES tt_item TYPE STANDARD TABLE OF ts_item.
    "-- Добавить шаг
    METHODS add_step
      IMPORTING
        iv_title      TYPE any DEFAULT 'Шаг'
      RETURNING
        VALUE(rr_ret) TYPE REF TO lcl_super_select_step.
    "-- Добавить текстовое поле значение которого берется из доменов по значению кода
    METHODS add_step_ddtext
      IMPORTING
        iv_id             TYPE any
        iv_text           TYPE any OPTIONAL
        iv_domain         TYPE any OPTIONAL
        iv_empty_as_empty TYPE wdy_boolean DEFAULT abap_false. " Пустое значение выводить как пусто без зхначения домена
    "-- Добавить поля с помощью выбора из таблицы
    METHODS add_step_select
      IMPORTING
        iv_from        TYPE any
        iv_select      TYPE any
        iv_keys        TYPE any
        iv_deleteempty TYPE any DEFAULT '' " Удалить строку если значение этого поля пустое
        iv_filter      TYPE wdy_boolean DEFAULT abap_undefined
      .
    "-- Получить указатель на шаг
    METHODS get_step
      IMPORTING
        iv_id         TYPE any
      RETURNING
        VALUE(rr_ret) TYPE REF TO lcl_super_select_step.
    "-- Получить указатель на шаг
    METHODS get_step_by_title
      IMPORTING
        iv_title      TYPE any
      RETURNING
        VALUE(rr_ret) TYPE REF TO lcl_super_select_step.
    "-- Установить подпрограмму обратного вызова, выполняемую перед запуском шага
    METHODS set_callback
      IMPORTING
        iv_callback   TYPE any
      RETURNING
        VALUE(rr_ret) TYPE REF TO lcl_super_select.
    "-- Останавлитваться перед выполнением шагов
    METHODS break_point_set
      IMPORTING
        iv_value      TYPE any
      RETURNING
        VALUE(rr_ret) TYPE REF TO lcl_super_select.
    "-- Останавлитваться перед выполнением шагов?
    METHODS break_point_get
      RETURNING
        VALUE(rv_ret) TYPE wdy_boolean.
    "-- Получить каталог ВСЕХ полей
    METHODS get_fcat
      RETURNING
        VALUE(rt_ret) TYPE lvc_t_fcat.
    "-- Установить визуализатор
    METHODS viewer_set
      IMPORTING
        iv_filename TYPE any " Имя файла визуализатора
      .
    "-- Запустить визуализацию
    METHODS viewer_start
      IMPORTING
        iv_title         TYPE any
        iv_start_as_html TYPE wdy_boolean DEFAULT abap_true
      RETURNING
        VALUE(rv_ret)    TYPE string.
    "-- Запустить
    METHODS run
      CHANGING
        ct_fc         TYPE lvc_t_fcat
      RETURNING
        VALUE(rr_ret) TYPE REF TO data.
    "-- Критическая ошибка
    CLASS-METHODS show_error
      IMPORTING
        text TYPE any
        v1   TYPE any OPTIONAL
        v2   TYPE any OPTIONAL
        v3   TYPE any OPTIONAL
        v4   TYPE any OPTIONAL.
  PRIVATE SECTION.
    "-- Добавить поле в каталог полей
    METHODS _add_fcat
      IMPORTING
        is_fcat TYPE lvc_s_fcat.
    "-- Разобрать список полей
    METHODS _parse_fields
      IMPORTING
        iv_fields              TYPE any
        iv_empty_right_eq_left TYPE wdy_boolean DEFAULT abap_true
      CHANGING
        ct_ret                 TYPE tt_item.
    "-- Рассчитать
    METHODS _recalc.
    "-- Получить код динамической программы
    METHODS _get_code
      IMPORTING
        iv_repid     TYPE sy-repid
        iv_repid_dyn TYPE sy-repid
      CHANGING
        ct_code      TYPE tt_code.
    "-- Дополнительные подпрограммы
    DATA mt_code TYPE tt_code.
    "-- Флаг выполнения расчета
    DATA mv_recalc TYPE wdy_boolean VALUE abap_false.
    "-- Список шагов
    DATA mt_steps TYPE lcl_super_select_step=>tt_super_select2_step.
    "-- Данные дял визуализации
    DATA mt_viewer TYPE STANDARD TABLE OF string.
    "-- Список полей
    DATA mt_fcat TYPE lvc_t_fcat.
    "-- Список полей для быстрой проверки наличия такого поля
    DATA mt_fields TYPE HASHED TABLE OF string WITH UNIQUE KEY table_line.
    "-- Время выполнения
    DATA mv_time TYPE i.
    "-- Список шагов
    DATA mt_stepsinfo TYPE lcl_super_select_step=>tt_super_select2_step.
    "-- Поля для которых значения берутся из доменов
    TYPES:
      BEGIN OF ts_ddtext,
        id             TYPE fieldname,
        text           TYPE fieldname,
        domain         TYPE domname,
        empty_as_empty TYPE wdy_boolean,
      END OF ts_ddtext.
    TYPES tt_ddtext TYPE STANDARD TABLE OF ts_ddtext.
    DATA mt_ddtext TYPE tt_ddtext.
    DATA mr_step_ddtext TYPE REF TO lcl_super_select_step.
    "-- Останавливать программу перед выполнение программ
    DATA mv_break_point TYPE wdy_boolean VALUE abap_false.
    "-- Подпрогамма обратного вызова
    DATA mv_callback TYPE string.
ENDCLASS.


CLASS lcl_super_select_step IMPLEMENTATION.
  "-- Установить имя подпрограмм шага
  METHOD set_formname.
    mv_formname = iv_formname.
    "-- Вернуть указатель на себя
    rr_ret = me.
  ENDMETHOD.
  "-- Установить тип подпрограммы подпрограммы (ТАБЛИЧНЫЙ ТИП!)
  METHOD set_type.
    mv_type = iv_type.
    TRANSLATE mv_type TO UPPER CASE.
    "-- Псевдонимы
    DATA lt_aliases TYPE mr_select->tt_item.
    DATA ls_alias LIKE LINE OF lt_aliases.
    mr_select->_parse_fields(
        EXPORTING
          iv_fields = iv_alias
        CHANGING
          ct_ret = lt_aliases
     ).
    SORT lt_aliases BY right.
    "--
    DATA lr_table_descr TYPE REF TO cl_abap_tabledescr.
    DATA lr_struct_descr TYPE REF TO cl_abap_structdescr.
    TRY.
        lr_table_descr ?= cl_abap_tabledescr=>describe_by_name( mv_type ).
        lr_struct_descr ?= lr_table_descr->get_table_line_type( ).
      CATCH cx_sy_move_cast_error.
        CALL METHOD mr_select->show_error(
            text = 'Неизвестный (не табличный) тип данных [&1] шага [&2]'
            v1   = mv_type
            v2   = me->get_title( )
                   ).
    ENDTRY.
    FIELD-SYMBOLS <ls_component> TYPE abap_compdescr.
    DATA lr_field_descr TYPE REF TO cl_abap_datadescr.
    "-- Удалить список ключевых полей
    DATA lt_components LIKE lr_struct_descr->components[].
    lt_components[] = lr_struct_descr->components[].
    FIELD-SYMBOLS <ls_relation> LIKE LINE OF mt_relations.
    LOOP AT mt_relations ASSIGNING <ls_relation>.
      DELETE lt_components WHERE name = <ls_relation>-aliasform.
    ENDLOOP.
    "-- Это шаг коррекции?
    DATA lv_is_correction TYPE wdy_boolean.
    "-- Есть в структуре поле шага коррекции?
    READ TABLE lt_components TRANSPORTING NO FIELDS
      WITH KEY
        name = mc_correction_fieldname
    .
    IF sy-subrc = 0.
      "-- Удалить поле из списка полей
      DELETE lt_components INDEX sy-tabix.
      "-- Это шаг корректировки существующего поля
      lv_is_correction = abap_true.
    ELSE.
      "-- Это шаг расчета нового поля
      lv_is_correction = abap_false.
    ENDIF.
    "-- Добавить все поля с их типами
    LOOP AT lt_components ASSIGNING <ls_component>.
      lr_field_descr = lr_struct_descr->get_component_type( <ls_component>-name ).
      "-- Искать название поля в подпрограмме расчета
      READ TABLE  lt_aliases INTO ls_alias
        WITH KEY
          right = <ls_component>-name
        BINARY SEARCH
      .
      IF sy-subrc <> 0.
        ls_alias-left = <ls_component>-name.
      ENDIF.
      "-- Добавить поле
      IF lv_is_correction = abap_true.
        me->add_field_correction(
          iv_name = ls_alias-left
          iv_aliasform  = <ls_component>-name
        ).
      ELSE.
        me->add_field(
          iv_name = ls_alias-left
          iv_type = lr_field_descr->get_relative_name( )
          iv_aliasform  = <ls_component>-name
        ).
      ENDIF.
    ENDLOOP.
    "-- Вернуть указатель на себя
    rr_ret = me.
  ENDMETHOD.
  "-- Добавить зависимость
  METHOD add_relation.
    DATA lt_relations TYPE STANDARD TABLE OF string.
    SPLIT iv_relation AT ',' INTO TABLE lt_relations.
    FIELD-SYMBOLS <lv_relation> LIKE LINE OF lt_relations.
    DATA ls_relations LIKE LINE OF mt_relations.
    LOOP AT lt_relations ASSIGNING <lv_relation>.
      CLEAR ls_relations.
      ls_relations-name = <lv_relation>.
      TRANSLATE ls_relations-name TO UPPER CASE.
      "-- А может задан псевдоним для подпрограммы расчета шага?
      SPLIT ls_relations-name AT '=' INTO ls_relations-name ls_relations-aliasform.
      IF ls_relations-aliasform IS INITIAL.
        ls_relations-aliasform = ls_relations-name.
      ENDIF.
      APPEND ls_relations TO mt_relations.
    ENDLOOP.
    "-- Вернуть указатель на себя
    rr_ret = me.
  ENDMETHOD.
  "-- Название шага
  METHOD get_title.
    rv_ret = mv_title.
  ENDMETHOD.
  "-- Изменить флаг фильтра
  METHOD set_filter.
    mv_filter = iv_filter.
    "-- Вернуть указатель на себя
    rr_ret = me.
  ENDMETHOD.
  "-- Шаг активен?
  METHOD is_disable.
    rv_ret = mv_disable.
  ENDMETHOD.
  "-- Останавлитваться перед выполнением шага
  METHOD break_point_set.
    IF NOT ( iv_value IS INITIAL ).
      mv_break_point = abap_true.
    ELSE.
      mv_break_point = abap_false.
    ENDIF.
    "-- Вернуть указатель на себя
    rr_ret = me.
  ENDMETHOD.
  "-- Останавлитваться перед выполнением шага?
  METHOD break_point_get.
    rv_ret = mv_break_point.
  ENDMETHOD.
  "-- Скрывать данный шаг при визуализации
  METHOD set_hide.
    mv_hide = iv_hide.
    "-- Вернут ьуказатель на себя
    rr_ret = me.
  ENDMETHOD.
  "-- Запустить таймер с именем iv_name
  METHOD timer_start.
    DATA ls_timer LIKE LINE OF mt_timers.
    CLEAR ls_timer.
    ls_timer-name = iv_name.
    GET RUN TIME FIELD ls_timer-start.
    INSERT ls_timer INTO TABLE mt_timers.
  ENDMETHOD.
  "-- Остановить таймер с именем iv_name
  METHOD timer_end.
    FIELD-SYMBOLS <ls_timer> LIKE LINE OF mt_timers.
    DATA lv_name LIKE <ls_timer>-name.
    lv_name = iv_name.
    READ TABLE mt_timers ASSIGNING <ls_timer>
      WITH TABLE KEY
      name = lv_name.
    IF sy-subrc = 0.
      GET RUN TIME FIELD <ls_timer>-end.
      <ls_timer>-time = <ls_timer>-end - <ls_timer>-start.
    ENDIF.
  ENDMETHOD.
  "-- Добавить поле
  METHOD add_field.
    DATA ls_field LIKE LINE OF mt_fields.
    CLEAR ls_field.
    ls_field-name = iv_name. TRANSLATE ls_field-name TO UPPER CASE.
    ls_field-type = iv_type. TRANSLATE ls_field-type TO UPPER CASE.
    "-- Имя поля в подпрограмме расчета
    IF iv_aliasform IS INITIAL.
      ls_field-aliasform = ls_field-name.
    ELSE.
      ls_field-aliasform = iv_aliasform. TRANSLATE ls_field-aliasform TO UPPER CASE.
    ENDIF.
    "-- Добавить в список полей
    APPEND ls_field TO mt_fields.
    "--
    DATA ls_fcat TYPE lvc_s_fcat.
    CLEAR ls_fcat.
    "-- Определить FC поля
    DATA lv_tabname TYPE dd02l-tabname.
    DATA lv_fieldname LIKE ls_fcat-fieldname.
    "-- Разобрать
    SPLIT ls_field-type AT '-' INTO lv_tabname lv_fieldname.
    IF NOT ( lv_fieldname IS INITIAL ).
      "-- Это ссылка на поле таблицы/структуры
      DATA lt_fcat TYPE lvc_t_fcat.
      CALL FUNCTION 'LVC_FIELDCATALOG_MERGE'
        EXPORTING
          i_structure_name = lv_tabname
        CHANGING
          ct_fieldcat      = lt_fcat.
      READ TABLE lt_fcat INTO ls_fcat
        WITH KEY
        fieldname = lv_fieldname
        .
      IF sy-subrc <> 0.
        CALL METHOD mr_select->show_error(
            text = 'Ошибка типа данных [&1] шага [&2]'
            v1   = ls_field-type
            v2   = me->get_title( )
                   ).
      ENDIF.
      "-- Установить ссылку на поле
      ls_fcat-ref_field = lv_fieldname.
    ELSE.
      "-- Это тип данных
      CASE ls_field-type.
        WHEN 'I' OR 'P'.
          ls_fcat-inttype = ls_field-type.
        WHEN 'STRING'.
          ls_fcat-inttype = 'g'.
        WHEN OTHERS.
          DATA lv_ddobjname TYPE ddobjname.
          DATA ls_dd04v TYPE dd04v.
          lv_ddobjname = ls_field-type.
          CALL FUNCTION 'DDIF_DTEL_GET'
            EXPORTING
              name          = lv_ddobjname
            IMPORTING
              dd04v_wa      = ls_dd04v
            EXCEPTIONS
              illegal_input = 1
              OTHERS        = 2.
          IF sy-subrc <> 0.
            CALL METHOD mr_select->show_error(
                text = 'Ошибка типа данных [&1] шага [&2]'
                v1   = ls_field-type
                v2   = me->get_title( )
                       ).
          ENDIF.
          MOVE-CORRESPONDING ls_dd04v TO ls_fcat.
          "-- Тексты (Выборка одной строки)
          SELECT reptext scrtext_s scrtext_m scrtext_l
             FROM dd04t
            UP TO 1 ROWS
             INTO (ls_fcat-reptext,
               ls_fcat-scrtext_s,
               ls_fcat-scrtext_m,
               ls_fcat-scrtext_l)
            WHERE rollname = ls_dd04v-rollname
            AND ddlanguage  = sy-langu
            AND as4local = 'A'
          .
          ENDSELECT.
      ENDCASE.
    ENDIF.
    "-- Имя поля
    ls_fcat-fieldname = ls_field-name.
    "-- Добавить поле в список полей отчета
    mr_select->_add_fcat( ls_fcat ).
  ENDMETHOD.
  "-- Добавить поле для коррекции
  METHOD add_field_correction.
*        iv_name      TYPE any
*        iv_aliasform TYPE any OPTIONAL " Имя поля в подпрограмме расчета
    DATA ls_field_correction LIKE LINE OF mt_fields_correction.
    CLEAR ls_field_correction.
    ls_field_correction-name = iv_name. TRANSLATE ls_field_correction-name TO UPPER CASE.
    "-- Имя поля в подпрограмме расчета
    IF iv_aliasform IS INITIAL.
      ls_field_correction-aliasform = ls_field_correction-name.
    ELSE.
      ls_field_correction-aliasform = iv_aliasform. TRANSLATE ls_field_correction-aliasform TO UPPER CASE.
    ENDIF.
    "-- Добавить в список полей
    APPEND ls_field_correction TO mt_fields_correction.
  ENDMETHOD.
  "-- Искать в родительских шагах шаг ir_step
  METHOD _find_parent.
    rr_ret = abap_false.
    "--
    FIELD-SYMBOLS <lr_parent> LIKE LINE OF mt_parents.
    LOOP AT mt_parents ASSIGNING <lr_parent>.
      IF <lr_parent> = ir_step.
        rr_ret = abap_true.
      ELSE.
        "-- Проверим на зацикливание
        FIELD-SYMBOLS <lv_history> LIKE LINE OF it_history.
        READ TABLE it_history TRANSPORTING NO FIELDS
          WITH KEY
            table_line = <lr_parent>.
        IF sy-subrc = 0.
          DATA lv_str TYPE char1024.
          CLEAR lv_str.
          LOOP AT it_history ASSIGNING <lv_history>.
            lv_str = lv_str && `,[` && <lv_history>->get_title( ) && `]`.
          ENDLOOP.
          CALL METHOD mr_select->show_error(
              text = 'Циклическая зависимость шага [&1]:&2'
              v1   = <lr_parent>->get_title( )
              v2   = lv_str+1
                     ).
        ENDIF.
        DATA lt_history TYPE tt_super_select2_step.
        lt_history[] = it_history[].
        APPEND <lr_parent> TO lt_history.
        rr_ret = <lr_parent>->_find_parent( EXPORTING
          ir_step = ir_step
          it_history = lt_history
        ).
      ENDIF.
      "-- Если нашли, то дальше можно не продолжать
      IF rr_ret = abap_true.
        EXIT.
      ENDIF.
    ENDLOOP.
  ENDMETHOD.
ENDCLASS.

CLASS lcl_super_select IMPLEMENTATION.
  "-- Добавить шаг
  METHOD add_step.
    "-- Создать шаг
    CREATE OBJECT rr_ret.
    "-- Родительский генератор запроса
    rr_ret->mr_select = me.
    "-- Имя шага
    rr_ret->mv_title = iv_title.
    "-- Добавить шаг в список шагов расчета
    APPEND rr_ret TO mt_steps.
    "-- Уникальный идентификатор
    rr_ret->mv_uid = lines( mt_stepsinfo ) + 1.
    CONDENSE rr_ret->mv_uid.
    APPEND rr_ret TO mt_stepsinfo.
  ENDMETHOD.
  "-- Добавить текстовое поле значение которого берется из доменов по значению кода
  METHOD add_step_ddtext.
    "--
    DATA ls_ddtext LIKE LINE OF mt_ddtext.
    CLEAR ls_ddtext.
    ls_ddtext-id = iv_id. TRANSLATE ls_ddtext-id TO UPPER CASE.
    ls_ddtext-text = iv_text. TRANSLATE ls_ddtext-text TO UPPER CASE.
    ls_ddtext-domain = iv_domain. TRANSLATE ls_ddtext-domain TO UPPER CASE.
    ls_ddtext-empty_as_empty = iv_empty_as_empty.
    APPEND ls_ddtext TO mt_ddtext.
    "--
    IF mr_step_ddtext IS INITIAL.
      mr_step_ddtext = me->add_step( '@ddTEXT' ).
      mr_step_ddtext->set_formname( '__init_ddtext' ).
    ENDIF.
    "--
    mr_step_ddtext->add_relation( iv_id ).
    mr_step_ddtext->add_field(
      iv_name = iv_text
      iv_type = 'dd07t-ddtext'
    ).
  ENDMETHOD.
  METHOD add_step_select.
    "-- Определить название шага
    DATA lv_title TYPE string.
    lv_title = `@` && iv_from.
    TRANSLATE lv_title TO UPPER CASE.
    "-- Создать шаг
    DATA lr_step TYPE REF TO lcl_super_select_step.
    lr_step = me->add_step( lv_title ).
    "-- Установить имя подпрограммы
    lr_step->set_formname( `__` && lr_step->mv_uid ).
    "-- Установить тип
    lr_step->mv_type = `__tt_step_` && lr_step->mv_uid.
    TRANSLATE lr_step->mv_type TO UPPER CASE.
    "-- Выбираемые поля
    DATA lt_selects TYPE tt_item.
    FIELD-SYMBOLS <ls_select> LIKE LINE OF lt_selects.
    me->_parse_fields(
        EXPORTING
          iv_fields = iv_select
        CHANGING
          ct_ret = lt_selects
     ).
    "-- Добавить поля
    LOOP AT lt_selects ASSIGNING <ls_select>.
      lr_step->add_field(
        iv_name = <ls_select>-right
        iv_aliasform = <ls_select>-left
        iv_type = iv_from && `-` && <ls_select>-left
      ).
    ENDLOOP.

    "-- Ключевые поля
    DATA lt_keys TYPE tt_item.
    FIELD-SYMBOLS <ls_key> LIKE LINE OF lt_keys.
    me->_parse_fields(
        EXPORTING
          iv_fields = iv_keys
        CHANGING
          ct_ret = lt_keys
     ).
    "-- Ключевые поля тоже выбирать (но не константы)
    DATA lv_str TYPE string.
    FIELD-SYMBOLS <fs> TYPE any.
    LOOP AT lt_keys ASSIGNING <ls_key>.
      IF <ls_key>-right(1) <> '@'.
        APPEND <ls_key> TO lt_selects.
      ENDIF.
      "-- Проверить - это глобальная переменная?
      IF <ls_key>-right(1) = '@'.
        "-- А может нет такого параметра/глобального значения?
        CONCATENATE `(` sy-repid `)` <ls_key>-right+1 INTO lv_str.
        ASSIGN (lv_str) TO <fs>.
        CHECK sy-subrc = 0.
        "-- А может это поле из SY?
        lv_str = <ls_key>-right+1.
        IF strlen( lv_str ) > 3.
          TRANSLATE lv_str TO UPPER CASE.
          CHECK lv_str(3) <> 'SY-'.
        ENDIF.
      ENDIF.
      "-- Добавить связь
      lr_step->add_relation( <ls_key>-right && `=` && <ls_key>-left ).
    ENDLOOP.
    "-- Фильтр
    IF iv_filter <> abap_undefined.
      lr_step->set_filter( iv_filter ).
    ENDIF.

    APPEND `"-- Выборка полей из таблицы ` && iv_from TO mt_code.
    APPEND `TYPES:` TO mt_code.
    APPEND `  BEGIN OF __ts_step_` && lr_step->mv_uid && `,` TO mt_code.
    LOOP AT lt_selects ASSIGNING <ls_select>.
      APPEND `    ` && <ls_select>-left && `     TYPE ` && iv_from && `-` && <ls_select>-left && `,` TO mt_code.
    ENDLOOP.
    "append `    kostl_txt TYPE cskt-ltext,` to mt_code.
    APPEND `  END OF __ts_step_` && lr_step->mv_uid && `.` TO mt_code.
    APPEND `TYPES ` && lr_step->mv_type &&  ` type STANDARD table OF __ts_step_` && lr_step->mv_uid && `.` TO mt_code.
    APPEND `FORM ` && lr_step->mv_formname TO mt_code.
    APPEND `       CHANGING` TO mt_code.
    APPEND `         ct_` && iv_from &&  `_lite TYPE ` && lr_step->mv_type &&  `.` TO mt_code.
    "-- А может нужно удалить строки с пустыми ключевыми полями?
    IF NOT ( iv_deleteempty IS INITIAL ).
      DATA it_fields TYPE STANDARD TABLE OF string.
      FIELD-SYMBOLS <lv_fields> LIKE LINE OF it_fields.
      SPLIT iv_deleteempty AT ',' INTO TABLE it_fields.
      LOOP AT it_fields ASSIGNING <lv_fields>.
        APPEND `  DELETE ct_` && iv_from &&  `_lite WHERE ` && <lv_fields> && ` IS INITIAL.` TO mt_code.
      ENDLOOP.
      APPEND `  CHECK NOT ( ct_` && iv_from &&  `_lite[] IS INITIAL ).` TO mt_code.
    ENDIF.
    "--
    APPEND `  SELECT` TO mt_code.
    LOOP AT lt_selects ASSIGNING <ls_select>.
      APPEND `      ` && <ls_select>-left TO mt_code.
    ENDLOOP.
    APPEND `    FROM ` && iv_from TO mt_code.
    APPEND `    INTO CORRESPONDING FIELDS OF TABLE ct_` && iv_from &&  `_lite` TO mt_code.
    APPEND `     FOR ALL ENTRIES IN ct_` && iv_from &&  `_lite` TO mt_code.

    DATA lv_str2 TYPE string.
    LOOP AT lt_keys ASSIGNING <ls_key>.
      IF sy-tabix = 1.
        lv_str = 'WHERE'.
      ELSE.
        lv_str = '  AND'.
      ENDIF.
      IF <ls_key>-right(1) = '@'.
        lv_str2 = <ls_key>-right+1.
      ELSE.
        lv_str2 = `ct_` && iv_from &&  `_lite-` && <ls_key>-left.
      ENDIF.
      APPEND `   ` && lv_str && ` ` && <ls_key>-left && ` = ` && lv_str2 TO mt_code.
    ENDLOOP.
    APPEND `  .` TO mt_code.

    APPEND `ENDFORM.` TO mt_code.
  ENDMETHOD.
  "-- Получить указатель на шаг
  METHOD get_step.
    FIELD-SYMBOLS <lv_stepsinfo> LIKE LINE OF mt_stepsinfo.
    DATA lv_tabix TYPE sy-tabix.
    lv_tabix = iv_id.
    READ TABLE mt_stepsinfo ASSIGNING <lv_stepsinfo> INDEX lv_tabix.
    IF sy-subrc = 0.
      rr_ret = <lv_stepsinfo>.
    ELSE.
      CLEAR rr_ret.
    ENDIF.
  ENDMETHOD.
  "-- Получить указатель на шаг
  METHOD get_step_by_title.
    CLEAR rr_ret.
    "--
    DATA lv_title TYPE string.
    lv_title = iv_title.
    TRANSLATE lv_title TO UPPER CASE.
    FIELD-SYMBOLS <lv_step> LIKE LINE OF mt_steps.
    DATA lv_title_step TYPE string.
    LOOP AT mt_steps ASSIGNING <lv_step>.
      lv_title_step = <lv_step>->get_title( ).
      TRANSLATE lv_title_step TO UPPER CASE.
      IF lv_title_step = lv_title.
        rr_ret = <lv_step>.
        EXIT.
      ENDIF.
    ENDLOOP.
  ENDMETHOD.
  "-- Установит ьподпрограмму обратного вызова, выполняемую перед запуском шага
  METHOD set_callback.
    mv_callback = iv_callback.
    "-- Вернуть указатель на себя
    rr_ret = me.
  ENDMETHOD.
  "-- Останавлитваться перед выполнением шагов
  METHOD break_point_set.
    IF NOT ( iv_value IS INITIAL ).
      mv_break_point = abap_true.
    ELSE.
      mv_break_point = abap_false.
    ENDIF.
    "-- Вернуть указатель на себя
    rr_ret = me.
  ENDMETHOD.
  "-- Останавлитваться перед выполнением шагов?
  METHOD break_point_get.
    rv_ret = mv_break_point.
  ENDMETHOD.
  "-- Получить каталог ВСЕХ полей
  METHOD get_fcat.
    rt_ret[] = mt_fcat[].
  ENDMETHOD.
  "-- Установить визуализатор
  METHOD viewer_set.
    REFRESH mt_viewer.
    "-- Читать файл
    DATA lv_filename TYPE string.
    lv_filename = iv_filename.
    CALL METHOD cl_gui_frontend_services=>gui_upload
      EXPORTING
        filename = lv_filename
        codepage = '4110' " UTF-8
      CHANGING
        data_tab = mt_viewer
      EXCEPTIONS
        OTHERS   = 1.
    IF sy-subrc <> 0.
      REFRESH mt_viewer.
    ENDIF.
  ENDMETHOD.
  "-- Запустить визуализацию
  METHOD viewer_start.
    "-- Визуализатор загружен?
    CHECK NOT ( mt_viewer[] IS INITIAL ).

    "-- Рассчитать
    CALL METHOD me->_recalc( ).

    "-- Сформировать JSON данные
    DATA lv_json TYPE string.
    DATA lv_str TYPE string.
    "--
    lv_json = lv_json && 'var data_arbor = {' && cl_abap_char_utilities=>newline.
    "-- Время работы
    lv_str = mv_time.
    CONDENSE lv_str.
    lv_json = lv_json && `  "time": ` && lv_str && `,` && cl_abap_char_utilities=>newline.
    "-- Шаги
    FIELD-SYMBOLS <lr_step> LIKE LINE OF me->mt_steps.
    FIELD-SYMBOLS <ls_field> LIKE LINE OF <lr_step>->mt_fields.
    FIELD-SYMBOLS <ls_field_correction> LIKE LINE OF <lr_step>->mt_fields_correction.
    FIELD-SYMBOLS <ls_relation> LIKE LINE OF <lr_step>->mt_relations.
    lv_json = lv_json && '  "nodes": [' && cl_abap_char_utilities=>newline.
    LOOP AT mt_steps ASSIGNING <lr_step>.
      lv_json = lv_json && '    {'.
      " {name: "node_1",time: 123,fields : 'a,b',rels:'@s_bukrs',filter:1,transform:0,type:"ts_anla"},
      lv_json = lv_json && ' "name" : "' && <lr_step>->mv_uid && '"'.
      lv_json = lv_json && ', "title" : "' && <lr_step>->get_title( ) && '"'.
      IF <lr_step>->mv_filter = abap_true.
        lv_json = lv_json && ', "filter" : 1'.
      ENDIF.
      IF <lr_step>->mv_disable = abap_true.
        lv_json = lv_json && ', "disable" : 1'.
      ENDIF.
      IF <lr_step>->mv_hide = abap_true.
        lv_json = lv_json && ', "hide" : 1'.
      ENDIF.
      IF NOT ( <lr_step>->mt_fields_correction[] IS INITIAL ).
        lv_json = lv_json && ', "correction" : 1'.
      ENDIF.
      lv_json = lv_json && ', "type" : "' && <lr_step>->mv_type && '"'.
      "--
      CLEAR lv_str.
      LOOP AT <lr_step>->mt_fields ASSIGNING <ls_field>.
        lv_str = lv_str && ',' && <ls_field>-name.
      ENDLOOP.
      LOOP AT <lr_step>->mt_fields_correction ASSIGNING <ls_field_correction>.
        lv_str = lv_str && ',' && <ls_field_correction>-name.
      ENDLOOP.
      IF NOT ( lv_str IS INITIAL ).
        lv_str = lv_str+1.
      ENDIF.
      lv_json = lv_json && ', "fields" : "' && lv_str && '"'.
      "--
      CLEAR lv_str.
      LOOP AT <lr_step>->mt_relations ASSIGNING <ls_relation>.
        lv_str = lv_str && ',' && <ls_relation>-name.
      ENDLOOP.
      IF NOT ( lv_str IS INITIAL ).
        lv_json = lv_json && ', "relations" : "' && lv_str+1 && '"'.
      ENDIF.
      "--
      lv_json = lv_json && `, "time" : {`.
      LOOP AT <lr_step>->mt_timers ASSIGNING FIELD-SYMBOL(<ls_timer>).
        lv_str = <ls_timer>-time.
        CONDENSE lv_str.
        lv_json = lv_json && ` "` && <ls_timer>-name && `" : ` && lv_str && `,`.
      ENDLOOP.
      lv_json = lv_json && ` }`.
      "--
      lv_json = lv_json && ' },' && cl_abap_char_utilities=>newline.
    ENDLOOP.
    lv_json = lv_json && '  ],' && cl_abap_char_utilities=>newline.
    "--
    lv_json = lv_json && '  "edges": [' && cl_abap_char_utilities=>newline.
    LOOP AT mt_steps ASSIGNING <lr_step>.
      "{src: "node_3", dest: "node_2"},
      LOOP AT <lr_step>->mt_parents ASSIGNING FIELD-SYMBOL(<lr_parent>).
        lv_json = lv_json && '    { "src": "' && <lr_step>->mv_uid && '", "dest": "' && <lr_parent>->mv_uid && '"},' && cl_abap_char_utilities=>newline.
      ENDLOOP.
    ENDLOOP.
    lv_json = lv_json && '  ],' && cl_abap_char_utilities=>newline.
    "--
    lv_json = lv_json && cl_abap_char_utilities=>newline && '};'.
    "-- Сформировать код визуализатора
    DATA lt_lines LIKE mt_viewer[].
    lt_lines[] = mt_viewer[].
    "-- Заменить заголовок
    REPLACE ALL OCCURRENCES OF '{{ TITLE }}'
     IN TABLE lt_lines WITH iv_title
     IGNORING CASE.
    "--
    lv_json = '<script type="text/javascript">' && lv_json && '</script>'.
    "-- Вставить в документ данные для визуализации
    REPLACE ALL OCCURRENCES OF '<script src="data.js"></script>'
     IN TABLE lt_lines WITH lv_json
     IGNORING CASE.


    IF iv_start_as_html = abap_true.

      "-- Сформировать временный файл
      DATA lv_filename_temp TYPE string.
      cl_gui_frontend_services=>get_temp_directory(
      CHANGING
        temp_dir = lv_filename_temp
      ).
      cl_gui_cfw=>flush( ).

      "-- Сгенерировать случайное число
      lv_str = cl_abap_random_packed=>create( seed = cl_abap_random=>seed( )  min = 0 max = 999999999999999 )->get_next( ).
      CONDENSE lv_str.
      lv_filename_temp = lv_filename_temp && '\' && sy-repid && '_' && sy-datum && '_' && sy-uzeit && '_' && lv_str && '.html'.

      "-- Сохранить файл
      CALL METHOD cl_gui_frontend_services=>gui_download
        EXPORTING
*         bin_filesize            =
          filename                = lv_filename_temp
*         filetype                = 'ASC'
*         append                  = SPACE
*         write_field_separator   = SPACE
*         header                  = '00'
*         trunc_trailing_blanks   = SPACE
*         write_lf                = 'X'
*         col_select              = SPACE
*         col_select_mask         = SPACE
*         dat_mode                = SPACE
*         confirm_overwrite       = SPACE
*         no_auth_check           = SPACE
          codepage                = '4110'
*         ignore_cerr             = ABAP_TRUE
*         replacement             = '#'
*         write_bom               = SPACE
*         trunc_trailing_blanks_eol = 'X'
*         wk1_n_format            = SPACE
*         wk1_n_size              = SPACE
*         wk1_t_format            = SPACE
*         wk1_t_size              = SPACE
*         show_transfer_status    = 'X'
*         fieldnames              =
*         write_lf_after_last_line  = 'X'
*         virus_scan_profile      = '/SCET/GUI_DOWNLOAD'
*  IMPORTING
*         filelength              =
        CHANGING
          data_tab                = lt_lines
        EXCEPTIONS
          file_write_error        = 1
          no_batch                = 2
          gui_refuse_filetransfer = 3
          invalid_type            = 4
          no_authority            = 5
          unknown_error           = 6
          header_not_allowed      = 7
          separator_not_allowed   = 8
          filesize_not_allowed    = 9
          header_too_long         = 10
          dp_error_create         = 11
          dp_error_send           = 12
          dp_error_write          = 13
          unknown_dp_error        = 14
          access_denied           = 15
          dp_out_of_memory        = 16
          disk_full               = 17
          dp_timeout              = 18
          file_not_found          = 19
          dataprovider_exception  = 20
          control_flush_error     = 21
          not_supported_by_gui    = 22
          error_no_gui            = 23
          OTHERS                  = 24.
      CHECK sy-subrc = 0.

      "-- Открыть файл для просмотра
      CALL METHOD cl_gui_frontend_services=>execute
        EXPORTING
          document = lv_filename_temp
        EXCEPTIONS
          OTHERS   = 1.
    ENDIF.
  ENDMETHOD.
  "-- Критическая ошибка
  METHOD show_error.
    DATA lv_text TYPE string.
    lv_text = text.
    REPLACE ALL OCCURRENCES OF `&1` IN lv_text WITH v1.
    REPLACE ALL OCCURRENCES OF `&2` IN lv_text WITH v2.
    REPLACE ALL OCCURRENCES OF `&3` IN lv_text WITH v3.
    REPLACE ALL OCCURRENCES OF `&4` IN lv_text WITH v4.
    "-- Выпасть в дамп с сообщением
    MESSAGE lv_text TYPE 'X'.
  ENDMETHOD.
  "-- Добавить поле в каталог полей
  METHOD _add_fcat.
    DATA ls_fcat TYPE lvc_s_fcat.
    ls_fcat = is_fcat.
    "-- Затереть индекс поля
    CLEAR ls_fcat-col_pos.
    "-- Поле по умолчанию видимо
    CLEAR ls_fcat-no_out.
    "-- Добавить в список
    APPEND ls_fcat TO mt_fcat.
    "-- Проверим: а может поле добавляют второй раз?
    DATA lv_field LIKE LINE OF mt_fields.
    lv_field = ls_fcat-fieldname.
    READ TABLE mt_fields TRANSPORTING NO FIELDS
      WITH TABLE KEY
        table_line = lv_field
    .
    IF sy-subrc = 0.
      me->show_error(
        text = 'Поле [&1] добавляется второй раз!'
        v1 = lv_field
      ).
    ELSE.
      INSERT lv_field INTO TABLE mt_fields.
    ENDIF.
  ENDMETHOD.
  "-- Разобрать список полей
  METHOD _parse_fields.
    REFRESH ct_ret.
    DATA lt_fields TYPE STANDARD TABLE OF string.
    SPLIT iv_fields AT ',' INTO TABLE lt_fields.
    FIELD-SYMBOLS <lv_fields> LIKE LINE OF lt_fields.
    DATA ls_ret LIKE LINE OF ct_ret.
    LOOP AT lt_fields ASSIGNING <lv_fields>.
      CLEAR ls_ret.
      TRANSLATE <lv_fields> TO UPPER CASE.
      SPLIT <lv_fields> AT '=' INTO ls_ret-left ls_ret-right.
      IF iv_empty_right_eq_left = abap_true AND ls_ret-right IS INITIAL.
        ls_ret-right = ls_ret-left.
      ENDIF.
      INSERT ls_ret INTO TABLE ct_ret.
    ENDLOOP.
  ENDMETHOD.
  "-- Рассчитать
  METHOD _recalc.
    CHECK mv_recalc = abap_false.
    "-- Установить флаг что расчет был выполнен
    mv_recalc = abap_true.
    "--
    FIELD-SYMBOLS <lr_step> LIKE LINE OF mt_steps.
    FIELD-SYMBOLS <ls_field> LIKE LINE OF <lr_step>->mt_fields.
    FIELD-SYMBOLS <ls_field_correction> LIKE LINE OF <lr_step>->mt_fields_correction.
    FIELD-SYMBOLS <ls_relation> LIKE LINE OF <lr_step>->mt_relations.
    "-- Определить родительские шаги
    TYPES:
      BEGIN OF ts_field_for_step,
        name     LIKE <ls_field>-name,
        formname LIKE <lr_step>->mv_formname,
        step     TYPE REF TO lcl_super_select_step,
      END OF ts_field_for_step.
    TYPES tt_field_for_step TYPE STANDARD TABLE OF ts_field_for_step.
    DATA lt_field_for_step TYPE tt_field_for_step.
    DATA ls_field_for_step LIKE LINE OF lt_field_for_step.
    LOOP AT mt_steps ASSIGNING <lr_step>.
      CLEAR ls_field_for_step.
      ls_field_for_step-step = <lr_step>.
      "-- Группы, где определяются поля
      LOOP AT <lr_step>->mt_fields ASSIGNING <ls_field>.
        ls_field_for_step-name = <ls_field>-name.
        APPEND ls_field_for_step TO lt_field_for_step.
      ENDLOOP.
      "-- Группы, где корректируются поля
      LOOP AT <lr_step>->mt_fields_correction ASSIGNING <ls_field_correction>.
        ls_field_for_step-name = <ls_field_correction>-name.
        APPEND ls_field_for_step TO lt_field_for_step.
      ENDLOOP.
    ENDLOOP.
    SORT lt_field_for_step BY name.
    FIELD-SYMBOLS <ls_field_for_step> LIKE LINE OF lt_field_for_step.
    LOOP AT mt_steps ASSIGNING <lr_step>.
      "-- Зависимость от ключевых полей
      LOOP AT <lr_step>->mt_relations ASSIGNING <ls_relation>.
        "-- Ищем родительские шаги для поля
        READ TABLE lt_field_for_step ASSIGNING <ls_field_for_step>
          WITH KEY
          name = <ls_relation>-name
          BINARY SEARCH.
        IF sy-subrc = 0.
          LOOP AT lt_field_for_step ASSIGNING <ls_field_for_step> FROM sy-tabix.
            IF <ls_field_for_step>-name <> <ls_relation>-name.
              EXIT.
            ENDIF.
            APPEND <ls_field_for_step>-step TO <lr_step>->mt_parents.
          ENDLOOP.
        ELSE.
          me->show_error(
            text = 'Не могу найти поле [&1] для шага [&2]'
            v1 = <ls_relation>-name
            v2 = <lr_step>->get_title( )
          ).
        ENDIF.
      ENDLOOP.
      "-- Зависимость от групп в которых поле корректируется
      LOOP AT <lr_step>->mt_fields_correction ASSIGNING <ls_field_correction>.
        "-- Ищем родительские шаги для поля
        READ TABLE lt_field_for_step ASSIGNING <ls_field_for_step>
          WITH KEY
          name = <ls_field_correction>-name
          BINARY SEARCH.
        IF sy-subrc = 0.
          LOOP AT lt_field_for_step ASSIGNING <ls_field_for_step> FROM sy-tabix.
            IF <ls_field_for_step>-name <> <ls_field_correction>-name.
              EXIT.
            ENDIF.
            "-- Зависимости от самого себя у шага быть не может!
            IF <ls_field_for_step>-step <> <lr_step>.
              APPEND <ls_field_for_step>-step TO <lr_step>->mt_parents.
            ENDIF.
          ENDLOOP.
        ELSE.
          me->show_error(
            text = 'Не могу найти поле [&1] для шага [&2]'
            v1 = <ls_relation>-name
            v2 = <lr_step>->get_title( )
          ).
        ENDIF.
      ENDLOOP.

      "-- Оставить только уникальные значения
      SORT <lr_step>->mt_parents.
      DELETE ADJACENT DUPLICATES FROM <lr_step>->mt_parents.
    ENDLOOP.

    "-- Сортируем все шаги чтобы в начале стояли
    " 1. Родительские шаги
    " 2. Фильтрующие шиги
    " 3. Швги которые имеют больше параметров селекционного экрана
    FIELD-SYMBOLS <lr_step1> LIKE <lr_step>.
    FIELD-SYMBOLS <lr_step2> LIKE <lr_step>.
    DATA lr_step LIKE <lr_step>.
    DATA lv_tabix TYPE sy-tabix.
    DATA lv_swap TYPE wdy_boolean.
    LOOP AT mt_steps ASSIGNING <lr_step1>.
      lv_tabix = sy-tabix + 1.
      LOOP AT mt_steps ASSIGNING <lr_step2> FROM lv_tabix.
        "-- По умолчанию не нужно менять местами шаги
        lv_swap = abap_false.
        "-- Если 2 является родителем 1
        IF <lr_step1>->_find_parent( <lr_step2> ) = abap_true.
          "-- меняем местами шаги
          lv_swap = abap_true.
        ELSE.
          "-- Если 1 не является родителем 2
          IF <lr_step2>->_find_parent( <lr_step1> ) = abap_false.
            "-- то сравним фильтры
            IF <lr_step1>->mv_filter <> <lr_step2>->mv_filter.
              "-- Шаги не равны. Если второй шаг с фильтром (а первая получается без фильтра)
              IF <lr_step2>->mv_filter = abap_true.
                "-- меняем местами шаги
                lv_swap = abap_true.
              ENDIF.
            ENDIF.
          ENDIF.
        ENDIF.
        "-- Если нужно
        IF lv_swap = abap_true.
          "-- то поменять
          lr_step = <lr_step2>.
          <lr_step2> = <lr_step1>.
          <lr_step1> = lr_step.
        ENDIF.
      ENDLOOP.
    ENDLOOP.
  ENDMETHOD.
  "-- Получить код динамической программы
  METHOD _get_code.
    "--
    REFRESH ct_code.
    "-- Читать код из программы
    READ REPORT iv_repid INTO ct_code.
    "--
    DATA lv_str TYPE char1024.
    CONCATENATE 'REPORT' iv_repid INTO lv_str SEPARATED BY space.
    DATA lv_str2 TYPE char1024.
    CONCATENATE 'REPORT' iv_repid_dyn INTO lv_str2 SEPARATED BY space.
    "--
    REPLACE ALL OCCURRENCES OF lv_str
       IN TABLE ct_code WITH lv_str2
       IGNORING CASE.

    "--
    FIELD-SYMBOLS <lr_step> LIKE LINE OF me->mt_steps.
    FIELD-SYMBOLS <ls_field> LIKE LINE OF <lr_step>->mt_fields.
    FIELD-SYMBOLS <ls_field_correction> LIKE LINE OF <lr_step>->mt_fields_correction.
    FIELD-SYMBOLS <ls_relation> LIKE LINE OF <lr_step>->mt_relations.

    "-- Инициализировать подпрограмму для определения текстов по доменам
    IF NOT ( mr_step_ddtext IS INITIAL ).
      DATA lt_fc TYPE lvc_t_fcat.
      FIELD-SYMBOLS <ls_fc> LIKE LINE OF lt_fc.
      lt_fc[] = mt_fcat[].
      SORT lt_fc BY fieldname.
      FIELD-SYMBOLS <ls_ddtext> LIKE LINE OF mt_ddtext.
      LOOP AT mt_ddtext ASSIGNING <ls_ddtext>.
        IF <ls_ddtext>-domain IS INITIAL.
          READ TABLE lt_fc ASSIGNING <ls_fc>
            WITH KEY
              fieldname = <ls_ddtext>-id
            BINARY SEARCH.
          IF sy-subrc = 0.
            <ls_ddtext>-domain = <ls_fc>-domname.
          ENDIF.
        ENDIF.
      ENDLOOP.

      APPEND `"-- Системный шаг для определения текстов по доменам` TO ct_code.
      APPEND `FORM ` && mr_step_ddtext->mv_formname && ` using` TO ct_code.
      APPEND `         ct_table type any table.` TO ct_code.
      APPEND `  TYPES:` TO ct_code.
      APPEND `    BEGIN OF ts_dd07_lite,` TO ct_code.
      APPEND `      domname    TYPE dd07l-domname,` TO ct_code.
      APPEND `      domvalue_l TYPE dd07l-domvalue_l,` TO ct_code.
      APPEND `      ddtext     TYPE dd07t-ddtext,` TO ct_code.
      APPEND `    END OF ts_dd07_lite.` TO ct_code.
      APPEND `  DATA lt_dd07_lite TYPE HASHED TABLE OF ts_dd07_lite WITH UNIQUE KEY domname domvalue_l.` TO ct_code.
      APPEND `  FIELD-SYMBOLS <ls_dd07_lite> LIKE LINE OF lt_dd07_lite.` TO ct_code.
      APPEND `  SELECT dd07l~domname dd07l~domvalue_l dd07t~ddtext` TO ct_code.
      APPEND `    INTO TABLE lt_dd07_lite` TO ct_code.
      APPEND `    FROM dd07l` TO ct_code.
      APPEND `      LEFT JOIN dd07t ON dd07t~domname    = dd07l~domname` TO ct_code.
      APPEND `                     AND dd07t~ddlanguage = sy-langu` TO ct_code.
      APPEND `                     AND dd07t~as4local   = dd07l~as4local` TO ct_code.
      APPEND `                     AND dd07t~valpos     = dd07l~valpos` TO ct_code.
      APPEND `                     AND dd07t~as4vers    = dd07l~as4vers` TO ct_code.
      APPEND `    WHERE dd07l~domname IN (` TO ct_code.
      LOOP AT mt_ddtext ASSIGNING <ls_ddtext>.
        IF sy-tabix > 1.
          APPEND `        ,` TO ct_code.
        ENDIF.
        APPEND `          '` && <ls_ddtext>-domain && `'` TO ct_code.
      ENDLOOP.
      APPEND `      )` TO ct_code.
      APPEND `      AND dd07l~as4local    = 'A'.` TO ct_code.
      APPEND `  field-symbols <ls_table> type any.` TO ct_code.
      APPEND `  field-symbols <id> type any.` TO ct_code.
      APPEND `  field-symbols <text> type any.` TO ct_code.
      APPEND `  DATA lv_domvalue_l TYPE dd07l-domvalue_l.` TO ct_code.
      APPEND `  LOOP AT ct_table ASSIGNING <ls_table>.` TO ct_code.
      LOOP AT mt_ddtext ASSIGNING <ls_ddtext>.
        APPEND `    ASSIGN COMPONENT '` && <ls_ddtext>-id && `' OF STRUCTURE <ls_table> TO <id>.` TO ct_code.
        APPEND `    IF sy-subrc = 0.` TO ct_code.
        IF <ls_ddtext>-empty_as_empty = abap_true.
          APPEND `      IF not ( <id> is initial ).` TO ct_code.
        ENDIF.
        APPEND `      ASSIGN COMPONENT '` && <ls_ddtext>-text && `' OF STRUCTURE <ls_table> TO <text>.` TO ct_code.
        APPEND `      IF sy-subrc = 0.` TO ct_code.
        APPEND `        lv_domvalue_l = <id>.` TO ct_code.
        APPEND `         READ TABLE lt_dd07_lite ASSIGNING <ls_dd07_lite>` TO ct_code.
        APPEND `            WITH TABLE KEY` TO ct_code.
        APPEND `              domname    = '` && <ls_ddtext>-domain && `'` TO ct_code.
        APPEND `              domvalue_l = lv_domvalue_l.` TO ct_code.
        APPEND `          IF sy-subrc = 0.` TO ct_code.
        APPEND `            <text> = <ls_dd07_lite>-ddtext.` TO ct_code.
        APPEND `          ENDIF.` TO ct_code.
        APPEND `      ENDIF.` TO ct_code.
        IF <ls_ddtext>-empty_as_empty = abap_true.
          APPEND `      ENDIF.` TO ct_code.
        ENDIF.
        APPEND `    ENDIF.` TO ct_code.
      ENDLOOP.
      APPEND `  ENDLOOP.` TO ct_code.
      APPEND `ENDFORM.` TO ct_code.
    ENDIF.

    "-- Добавить дополнительный код
    APPEND LINES OF mt_code TO ct_code.

    "-- Сформировать подпрограмму запуска
    APPEND `FORM RUN using` TO ct_code.
    APPEND `    pr_super_select type ref to object` TO ct_code.
    APPEND `    pr_table type ref to data.` TO ct_code.
    "APPEND '  break its_shva.' TO ct_code.

    DATA lt_rfieldlist TYPE STANDARD TABLE OF  rfieldlist.
    CALL FUNCTION 'GET_GLOBAL_SYMBOLS'
      EXPORTING
        program   = iv_repid
      TABLES
        fieldlist = lt_rfieldlist.
    DELETE lt_rfieldlist WHERE NOT ( flitl IS INITIAL ).
    DELETE lt_rfieldlist WHERE name(1) = '%'.
    DELETE lt_rfieldlist WHERE name = 'SSCRFIELDS'.
    DELETE lt_rfieldlist WHERE name = 'SSCRTEXTS'.
    DELETE lt_rfieldlist WHERE name = 'SYST'.
    DELETE lt_rfieldlist WHERE name = 'SY'.
    DELETE lt_rfieldlist WHERE name = 'SCREEN'.
    DELETE lt_rfieldlist WHERE name = 'VARI'.
    DELETE lt_rfieldlist WHERE type = 'r'. " Ссылки на объекты
    APPEND `  "-- Инициализировать глобальные переменные` TO ct_code.
    APPEND `  FIELD-SYMBOLS <fs> TYPE any.` TO ct_code.
    LOOP AT lt_rfieldlist ASSIGNING FIELD-SYMBOL(<ls_rfieldlist>).
      "--
      REPLACE ALL OCCURRENCES OF `[]` IN <ls_rfieldlist>-name WITH space.
      "-- Для таблиц добавим к имени []
      IF <ls_rfieldlist>-type = 'h'.
        <ls_rfieldlist>-name = <ls_rfieldlist>-name && `[]`.
      ENDIF.
      "--
      APPEND `  ASSIGN ('(` && iv_repid && `)` && <ls_rfieldlist>-name && `') TO <fs>.` TO ct_code.
      APPEND `  IF sy-subrc = 0.` TO ct_code.
      APPEND `    ` && <ls_rfieldlist>-name && ` = <fs>.` TO ct_code.
      APPEND `  ENDIF.` TO ct_code.
    ENDLOOP.
    "-- А может нужно остановиться перед вызовом шагов?
    APPEND `  DATA lv_break TYPE wdy_boolean.` TO ct_code.
    APPEND `  CALL METHOD pr_super_select->('BREAK_POINT_GET')` TO ct_code.
    APPEND `        RECEIVING  rv_ret = lv_break.` TO ct_code.
    APPEND `  IF lv_break = abap_true.` TO ct_code.
    APPEND `    BREAK-POINT.` TO ct_code.
    APPEND `  ENDIF.` TO ct_code.
    "-- Вызвать все шаги
    APPEND `  "== Вызвать расчет всех шагов` TO ct_code.
    DATA lv_steps_count TYPE char32.
    lv_steps_count = lines( mt_steps ).
    CONDENSE lv_steps_count.
    DATA lv_num_step TYPE char32.
    LOOP AT mt_steps ASSIGNING <lr_step>.
      lv_num_step = sy-tabix.
      APPEND `  "------ ` && <lr_step>->get_title( ) TO ct_code.
      IF NOT ( mv_callback IS INITIAL ).
        CONDENSE lv_num_step.
        APPEND `  PERFORM ` && mv_callback && ` USING '` && <lr_step>->get_title( ) && `' ` && lv_num_step && ` ` && lv_steps_count && `.` TO ct_code.
      ENDIF.
      APPEND `  PERFORM _step_prepare_` && <lr_step>->mv_uid && ` using pr_super_select pr_table.` TO ct_code.
    ENDLOOP.
    "--
    APPEND 'ENDFORM.' TO ct_code.

    "-- Сформировать подпрограммы выполнения шагов
    DATA lt_relations LIKE <lr_step>->mt_relations[].
    LOOP AT mt_steps ASSIGNING <lr_step>.
      "-- Список полей (зависимых, ключевых)
      lt_relations[] = <lr_step>->mt_relations[].
      "--
      APPEND `FORM _step_prepare_` && <lr_step>->mv_uid && ` using` TO ct_code.
      APPEND `         pr_super_select type ref to object` TO ct_code.
      APPEND `         pr_table type ref to data.` TO ct_code.
      APPEND `  DATA lr_step TYPE REF TO object.` TO ct_code.
      APPEND `  data lv_ret type wdy_boolean.` TO ct_code.
      APPEND `  CALL METHOD pr_super_select->('GET_STEP')` TO ct_code.
      APPEND `        EXPORTING iv_id = '` && <lr_step>->mv_uid && `'` TO ct_code.
      APPEND `        RECEIVING  rr_ret = lr_step.` TO ct_code.
      APPEND `  CALL METHOD lr_step->('IS_DISABLE')` TO ct_code.
      APPEND `        RECEIVING  rv_ret = lv_ret.` TO ct_code.
      APPEND `  CHECK lv_ret = abap_false.` TO ct_code.
      APPEND `  CALL METHOD lr_step->('TIMER_START')` TO ct_code.
      APPEND `        EXPORTING iv_name = 'all'.` TO ct_code.
      APPEND `  FIELD-SYMBOLS <lt_table> TYPE STANDARD TABLE.` TO ct_code.
      APPEND `  ASSIGN pr_table->* TO <lt_table>.` TO ct_code.
      IF NOT ( <lr_step>->mt_parents[] IS INITIAL ).
        "-- то проверим чтобы таблица была не пустая
        APPEND `  CHECK NOT ( <lt_table>[] IS INITIAL ).` TO ct_code.
      ENDIF.
      IF NOT ( <lr_step>->mv_type IS INITIAL ).
        APPEND `  field-symbols <ls_table> type any.` TO ct_code.
        APPEND `  field-symbols <fs> type any.` TO ct_code.
        APPEND `  data lt_data type ` && <lr_step>->mv_type && `.` TO ct_code.
        APPEND `  data ls_data like line of lt_data.` TO ct_code.
        APPEND `  field-symbols <ls_data> like line of lt_data.` TO ct_code.
        "-- Если это не первый (корневой, root) шаг
        IF NOT ( <lr_step>->mt_parents[] IS INITIAL ).
          APPEND `  "-- Сформировать данные` TO ct_code.
          IF NOT ( <lr_step>->mt_fields_correction[] IS INITIAL ).
            APPEND `  DATA lv_tabix type sy-tabix.` TO ct_code.
          ENDIF.
          APPEND `  loop at <lt_table> assigning <ls_table>.` TO ct_code.
          IF NOT ( <lr_step>->mt_fields_correction[] IS INITIAL ).
            APPEND `  lv_tabix = sy-tabix.` TO ct_code.
          ENDIF.
          APPEND `    clear ls_data.` TO ct_code.
          IF NOT ( <lr_step>->mt_fields_correction[] IS INITIAL ).
            APPEND `    ls_data-` && lcl_super_select_step=>mc_correction_fieldname && ` = lv_tabix.` TO ct_code.
          ENDIF.
          "-- Инициализация полей
          LOOP AT lt_relations ASSIGNING <ls_relation>.
            APPEND `    ASSIGN COMPONENT '` && <ls_relation>-name && `' OF STRUCTURE <ls_table> TO <fs>.` TO ct_code.
            APPEND `    IF sy-subrc = 0.` TO ct_code.
            APPEND `      ls_data-` && <ls_relation>-aliasform && ` = <fs>.` TO ct_code.
            APPEND `    ENDIF.` TO ct_code.
          ENDLOOP.
          LOOP AT <lr_step>->mt_fields_correction ASSIGNING <ls_field_correction>.
            APPEND `    ASSIGN COMPONENT '` && <ls_field_correction>-name && `' OF STRUCTURE <ls_table> TO <fs>.` TO ct_code.
            APPEND `    IF sy-subrc = 0.` TO ct_code.
            APPEND `      ls_data-` && <ls_field_correction>-aliasform && ` = <fs>.` TO ct_code.
            APPEND `    ENDIF.` TO ct_code.
          ENDLOOP.
          APPEND `    append ls_data to lt_data.` TO ct_code.
          APPEND `  endloop.` TO ct_code.
          "-- Если нужно
          IF NOT ( lt_relations[] IS INITIAL ) AND <lr_step>->mt_fields_correction[] IS INITIAL.
            APPEND `  "-- Убрать дублирующие записи` TO ct_code.
            "-- сортировать по ключевым полям
            APPEND `  sort lt_data by` TO ct_code.
            LOOP AT lt_relations ASSIGNING <ls_relation>.
              APPEND `            ` && <ls_relation>-aliasform TO ct_code.
            ENDLOOP.
            APPEND `  .` TO ct_code.
            APPEND `  DELETE ADJACENT DUPLICATES FROM lt_data COMPARING` TO ct_code.
            LOOP AT lt_relations ASSIGNING <ls_relation>.
              APPEND `            ` && <ls_relation>-aliasform TO ct_code.
            ENDLOOP.
            APPEND `  .` TO ct_code.
          ENDIF.
        ENDIF.
      ENDIF.
      "--
      APPEND `  "-- Вызвать подпрограмму шага` TO ct_code.
      APPEND `  CALL METHOD lr_step->('TIMER_START')` TO ct_code.
      APPEND `        EXPORTING iv_name = 'work'.` TO ct_code.
      APPEND `  DATA lv_break TYPE wdy_boolean.` TO ct_code.
      APPEND `  CALL METHOD lr_step->('BREAK_POINT_GET')` TO ct_code.
      APPEND `        RECEIVING  rv_ret = lv_break.` TO ct_code.
      APPEND `  IF lv_break = abap_true.` TO ct_code.
      APPEND `    BREAK-POINT.` TO ct_code.
      APPEND `  ENDIF.` TO ct_code.
      IF NOT ( <lr_step>->mv_type IS INITIAL ).
        APPEND `  perform ` && <lr_step>->mv_formname && ` changing lt_data.` TO ct_code.
      ELSE.
        APPEND `  perform ` && <lr_step>->mv_formname && ` changing <lt_table>.` TO ct_code.
      ENDIF.
      APPEND `  CALL METHOD lr_step->('TIMER_END')` TO ct_code.
      APPEND `        EXPORTING iv_name = 'work'.` TO ct_code.
      "--
      IF NOT ( <lr_step>->mv_type IS INITIAL ).
        APPEND `  "-- Записать данные в отчет` TO ct_code.
        IF NOT ( <lr_step>->mt_fields_correction[] IS INITIAL ).
          APPEND `  LOOP AT lt_data ASSIGNING <ls_data>.` TO ct_code.
          APPEND `    READ TABLE <lt_table> ASSIGNING <ls_table> index <ls_data>-` && lcl_super_select_step=>mc_correction_fieldname && `.` TO ct_code.
          APPEND `    IF sy-subrc <> 0.` TO ct_code.
          APPEND `      CALL METHOD lcl_super_select=>show_error( 'Сбился UID строки для шага [` && <lr_step>->get_title( ) && `]' ).` TO ct_code.
          APPEND `    ENDIF.` TO ct_code.
          LOOP AT <lr_step>->mt_fields_correction ASSIGNING <ls_field_correction>.
            APPEND `    ASSIGN COMPONENT '` && <ls_field_correction>-name && `' OF STRUCTURE <ls_table> TO <fs>.` TO ct_code.
            APPEND `    IF sy-subrc = 0.` TO ct_code.
            APPEND `      <fs> = <ls_data>-` && <ls_field_correction>-aliasform && `.` TO ct_code.
            APPEND `    ENDIF.` TO ct_code.
          ENDLOOP.
          APPEND `  ENDLOOP.` TO ct_code.
        ELSE.
          IF NOT ( <lr_step>->mt_parents[] IS INITIAL ).
            "-- Если нужно
            IF NOT ( lt_relations[] IS INITIAL ).
              "-- то сортировать по ключевым полям
              APPEND `  sort lt_data by` TO ct_code.
              LOOP AT lt_relations ASSIGNING <ls_relation>.
                APPEND `            ` && <ls_relation>-aliasform TO ct_code.
              ENDLOOP.
              APPEND `  .` TO ct_code.
            ENDIF.
            APPEND `  loop at <lt_table> assigning <ls_table>.` TO ct_code.
            APPEND `    clear ls_data.` TO ct_code.
            "-- Инициализация полей
            LOOP AT lt_relations ASSIGNING <ls_relation>.
              APPEND `    ASSIGN COMPONENT '` && <ls_relation>-name && `' OF STRUCTURE <ls_table> TO <fs>.` TO ct_code.
              APPEND `    IF sy-subrc = 0.` TO ct_code.
              APPEND `      ls_data-` && <ls_relation>-aliasform && ` = <fs>.` TO ct_code.
              APPEND `    ENDIF.` TO ct_code.
            ENDLOOP.
            "-- Если нужно
            IF NOT ( lt_relations[] IS INITIAL ).
              "-- то сортировать по ключевым полям
              APPEND `    read table lt_data assigning <ls_data> with key` TO ct_code.
              LOOP AT lt_relations ASSIGNING <ls_relation>.
                APPEND `          ` && <ls_relation>-aliasform && ` = ls_data-` && <ls_relation>-aliasform TO ct_code.
              ENDLOOP.
              APPEND `    BINARY SEARCH.` TO ct_code.
              APPEND `    if sy-subrc = 0.` TO ct_code.
              LOOP AT <lr_step>->mt_fields ASSIGNING <ls_field>.
                APPEND `      ASSIGN COMPONENT '` && <ls_field>-name && `' OF STRUCTURE <ls_table> TO <fs>.` TO ct_code.
                APPEND `      IF sy-subrc = 0.` TO ct_code.
                APPEND `        <fs> = <ls_data>-` && <ls_field>-aliasform && `.` TO ct_code.
                APPEND `      ENDIF.` TO ct_code.
              ENDLOOP.
              "-- Если это фильтрующий шаг
              IF <lr_step>->mv_filter = abap_true.
                APPEND `    else.` TO ct_code.
                APPEND `      delete <lt_table>.` TO ct_code.
              ENDIF.
              APPEND `    endif.` TO ct_code.
            ENDIF.
            APPEND `  endloop.` TO ct_code.
          ELSE.
            APPEND `  LOOP AT lt_data ASSIGNING <ls_data>.` TO ct_code.
            APPEND `    APPEND INITIAL LINE TO <lt_table> ASSIGNING <ls_table>.` TO ct_code.
            LOOP AT <lr_step>->mt_fields ASSIGNING <ls_field>.
              APPEND `    ASSIGN COMPONENT '` && <ls_field>-name && `' OF STRUCTURE <ls_table> TO <fs>.` TO ct_code.
              APPEND `    IF sy-subrc = 0.` TO ct_code.
              APPEND `      <fs> = <ls_data>-` && <ls_field>-aliasform && `.` TO ct_code.
              APPEND `    ENDIF.` TO ct_code.
            ENDLOOP.
            APPEND `  ENDLOOP.` TO ct_code.
          ENDIF.
        ENDIF.
      ENDIF.
      APPEND `  CALL METHOD lr_step->('TIMER_END')` TO ct_code.
      APPEND `        EXPORTING iv_name = 'all'.` TO ct_code.
      APPEND `ENDFORM.` TO ct_code.
    ENDLOOP.
  ENDMETHOD.
  "-- Запустить
  METHOD run.
    CLEAR rr_ret.
    "--
    DATA lv_time_start TYPE timestamp.
    GET RUN TIME FIELD lv_time_start.

    "-- Рассчитать
    CALL METHOD me->_recalc( ).

    "-- Определить активность шагов
    FIELD-SYMBOLS <lr_step> LIKE LINE OF mt_steps.
    FIELD-SYMBOLS <ls_field> LIKE LINE OF <lr_step>->mt_fields.
    FIELD-SYMBOLS <ls_field_correction> LIKE LINE OF <lr_step>->mt_fields_correction.
    DATA lt_fc LIKE ct_fc[].
    lt_fc[] = ct_fc[].
    SORT lt_fc BY fieldname.
    DATA lv_fieldname TYPE lvc_s_fcat-fieldname.
    LOOP AT mt_steps ASSIGNING <lr_step>.
      IF <lr_step>->mv_filter = abap_true.
        <lr_step>->mv_disable = abap_false.
        CONTINUE.
      ENDIF.
      "-- Определить активность шага
      <lr_step>->mv_disable = abap_true.
      LOOP AT <lr_step>->mt_fields ASSIGNING <ls_field>.
        lv_fieldname = <ls_field>-name.
        READ TABLE lt_fc TRANSPORTING NO FIELDS
          WITH KEY
            fieldname = lv_fieldname
          BINARY SEARCH.
        IF sy-subrc = 0.
          "-- Если хоть одно поле шага должно быть в отчете,
          "   то значит шаг активен
          <lr_step>->mv_disable = abap_false.
          "-- дальше можно не продолжать
          EXIT.
        ENDIF.
      ENDLOOP.
      "-- А может это шаг коррекции?
      LOOP AT <lr_step>->mt_fields_correction ASSIGNING <ls_field_correction>.
        lv_fieldname = <ls_field_correction>-name.
        READ TABLE lt_fc TRANSPORTING NO FIELDS
          WITH KEY
            fieldname = lv_fieldname
          BINARY SEARCH.
        IF sy-subrc = 0.
          "-- Если хоть одно поле шага должно быть в отчете,
          "   то значит шаг активен
          <lr_step>->mv_disable = abap_false.
          "-- дальше можно не продолжать
          EXIT.
        ENDIF.
      ENDLOOP.
    ENDLOOP.
    "-- А теперь для активных шагов включим все родительские шаги
    DATA lt_steps LIKE mt_steps[].
    LOOP AT mt_steps ASSIGNING <lr_step>.
      INSERT <lr_step> INTO lt_steps INDEX 1.
    ENDLOOP.
    FIELD-SYMBOLS <lr_parent> LIKE LINE OF <lr_step>->mt_parents.
    LOOP AT lt_steps ASSIGNING <lr_step>.
      "-- Если шаг активен
      IF <lr_step>->mv_disable = abap_false.
        "-- то активировать все родительские группы
        LOOP AT <lr_step>->mt_parents ASSIGNING <lr_parent>.
          <lr_parent>->mv_disable = abap_false.
        ENDLOOP.
      ENDIF.
    ENDLOOP.
    "-- А теперь добавим в каталог поля все поля активных шагов (которых нет в списке)
    DATA ls_fc LIKE LINE OF ct_fc.
    DATA lt_fc_all LIKE ct_fc[].
    lt_fc_all[] = mt_fcat[].
    SORT lt_fc_all BY fieldname.
    LOOP AT lt_steps ASSIGNING <lr_step>.
      "-- Только активные шаги
      CHECK <lr_step>->mv_disable = abap_false.
      "--
      LOOP AT <lr_step>->mt_fields ASSIGNING <ls_field>.
        lv_fieldname = <ls_field>-name.
        READ TABLE lt_fc TRANSPORTING NO FIELDS
          WITH KEY
            fieldname = lv_fieldname
          BINARY SEARCH.
        IF sy-subrc <> 0.
          READ TABLE lt_fc_all INTO ls_fc
            WITH KEY
              fieldname = lv_fieldname
            BINARY SEARCH.
          IF sy-subrc = 0.
            ls_fc-no_out = 'X'. " Скрыть поле
            APPEND ls_fc TO ct_fc.
          ENDIF.
        ENDIF.
      ENDLOOP.
    ENDLOOP.

    "-- Создать таблицу данных отчета
    "break its_shva.
    CALL METHOD cl_alv_table_create=>create_dynamic_table
      EXPORTING
        it_fieldcatalog = ct_fc
      IMPORTING
        ep_table        = rr_ret.
    "-- Текущая программа
    DATA lv_repid TYPE sy-repid.
    lv_repid = sy-repid.
    "-- Динамическая программа
    DATA lv_repid_dyn TYPE sy-repid.
    CONCATENATE lv_repid '_DYN' INTO lv_repid_dyn.
    "--
    DATA lt_code TYPE tt_code.
    "-- Определим строку версий всех состовляющих прогрумма INCLUDE-ов
    DATA lv_vern LIKE LINE OF lt_code.
    CLEAR lv_vern.
    DATA lt_incltab TYPE STANDARD TABLE OF trdir-name.
    CALL FUNCTION 'RS_GET_ALL_INCLUDES'
      EXPORTING
        program      = lv_repid
      TABLES
        includetab   = lt_incltab
      EXCEPTIONS
        not_existent = 0
        no_program   = 0
        OTHERS       = 0.
    APPEND lv_repid TO lt_incltab.
    "-- Выберем версии всех программ
    SELECT name, vern
      FROM trdir
      INTO TABLE @DATA(lt_trdir_lite)
       FOR ALL ENTRIES IN @lt_incltab
     WHERE name = @lt_incltab-table_line.
    "-- Создадим строку версий всех программ в сортировке
    SORT lt_trdir_lite BY name.
    FIELD-SYMBOLS <ls_trdir_lite> LIKE LINE OF lt_trdir_lite.
    LOOP AT lt_trdir_lite ASSIGNING <ls_trdir_lite>.
      lv_vern = lv_vern && `,` && <ls_trdir_lite>-vern.
    ENDLOOP.
    lv_vern = `*-- ` && lv_vern+1.

    "-- Проверим: программа существует?
    DATA ls_trdir TYPE trdir.
    CLEAR ls_trdir.
    SELECT SINGLE *
      FROM trdir
      INTO ls_trdir
     WHERE name = lv_repid_dyn.
    IF sy-subrc = 0.
      "-- Читать код динимической программы
      READ REPORT lv_repid_dyn INTO lt_code.
      "-- Читать первую строку
      DATA lv_code LIKE LINE OF lt_code.
      READ TABLE lt_code INTO lv_code INDEX 1.
      "-- Сравним первую строку и код
      IF lv_vern <> lv_code.
        "-- Если не равны, то удалить динамическую прогрумма
        DELETE REPORT lv_repid_dyn.
        "--
        CLEAR ls_trdir.
      ENDIF.
    ENDIF.
    "-- Нужно генерировать программу?
    IF ls_trdir IS INITIAL.
      "-- Создадим код динамической программы ()
      CALL METHOD _get_code(
        EXPORTING
          iv_repid     = lv_repid
          iv_repid_dyn = lv_repid_dyn
        CHANGING
          ct_code      = lt_code[]
                         ).
      "-- Добавить в начало версии
      INSERT lv_vern INTO lt_code INDEX 1.
      "-- Добавить программу
      INSERT REPORT lv_repid_dyn FROM lt_code.
      IF sy-subrc = 0.
        COMMIT WORK AND WAIT.
      ENDIF.
    ENDIF.
    "-- Запустить выборки данных
    DATA lr_this TYPE REF TO object.
    lr_this = me.
    PERFORM ('RUN') IN PROGRAM (lv_repid_dyn) USING lr_this rr_ret.
    "--
    DATA lv_time_end TYPE timestamp.
    GET RUN TIME FIELD lv_time_end.
    mv_time = lv_time_end - lv_time_start.
  ENDMETHOD.
ENDCLASS.